package info.textgrid.lab.collatex.ui.parts;

import info.textgrid.lab.collatex.model.CollationResults;
import info.textgrid.lab.collatex.model.CollationResults.ICollationResultsListener;
import info.textgrid.lab.collatex.model.equivalences.Equivalences;
import info.textgrid.lab.collatex.ui.WitnessTransfer;
import info.textgrid.lab.collatex.ui.parts.CollationSetEditor.ICollationEventHandler;

import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import org.eclipse.jface.action.IMenuManager;
import org.eclipse.jface.action.IToolBarManager;
import org.eclipse.jface.viewers.DelegatingStyledCellLabelProvider;
import org.eclipse.jface.viewers.IStructuredContentProvider;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.Viewer;
import org.eclipse.nebula.jface.gridviewer.GridTableViewer;
import org.eclipse.nebula.jface.gridviewer.GridViewerColumn;
import org.eclipse.nebula.widgets.grid.Grid;
import org.eclipse.nebula.widgets.grid.GridColumn;
import org.eclipse.nebula.widgets.grid.GridDragSourceEffect;
import org.eclipse.swt.SWT;
import org.eclipse.swt.dnd.DND;
import org.eclipse.swt.dnd.DragSourceEvent;
import org.eclipse.swt.dnd.Transfer;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.ui.part.IPageSite;
import org.neo4j.graphdb.Transaction;

import com.google.common.collect.ImmutableSet;
import com.google.common.collect.ImmutableSet.Builder;
import com.google.common.collect.Maps;
import com.google.common.collect.RowSortedTable;

import eu.interedition.collatex.Token;
import eu.interedition.collatex.Witness;
import eu.interedition.collatex.graph.VariantGraph;
import eu.interedition.collatex.graph.VariantGraphTransposition;

/**
 * Uses a nebula {@link GridTableViewer} to display collation results in a
 * vertical alignment table: A column for each witness, a cell for each token.
 * This page tracks and visualizes the currently focused witness and renders
 * equivalent token cells appropriately. Cells can be dragged to an
 * {@link RegularisationEditor} to create {@link Equivalences}.
 */
public class GridResultPage extends AbstractResultPage implements ICollationEventHandler {

	@Override
	public void dispose() {
		if (collationSetEditor != null)
			collationSetEditor.removeCollationListener(this);
		super.dispose();
	}

	private Composite control;
	GridTableViewer gridTableViewer;
	private CollationResultsContentProvider contentProvider;
	CollationSetEditor collationSetEditor;

	/**
	 * Create the PageBookView Page.
	 */
	public GridResultPage() {
	}


	/**
	 * Create contents of the view part.
	 * 
	 * @param parent
	 */
	@Override
	public void createControl(final Composite parent) {
		final Composite container = new Composite(parent, SWT.NONE);
		container.setLayout(new GridLayout(1, false));
		{
			setViewer(new GridTableViewer(container, SWT.BORDER
					| SWT.V_SCROLL));
			final Grid grid = getViewer().getGrid();
			grid.setSelectionEnabled(false);
			grid.setLinesVisible(false);
			grid.setHeaderVisible(true);
			grid.setCellSelectionEnabled(true);
			grid.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, true, 1,
					1));

			contentProvider = new CollationResultsContentProvider();
			getViewer().setContentProvider(contentProvider);

			getViewer().addDragSupport(DND.DROP_COPY | DND.DROP_MOVE
					| DND.DROP_LINK,
					new Transfer[] { WitnessTransfer.getInstance() },
					new GridDragSourceEffect(grid) {

				@Override
				public void dragSetData(final DragSourceEvent event) {
					System.err.println("Drag data ..."); //$NON-NLS-1$
					if (event.data != null) {
						System.err.println("...already set"); //$NON-NLS-1$
						return;
					}
					final GridColumn focusColumn = grid.getColumns()[grid
					                                                 .getFocusCell().x];
					final Witness witness = contentProvider
							.getWitness(focusColumn);

					final IStructuredSelection selection = (IStructuredSelection) getViewer()
							.getSelection();
					final Builder<Token> builder = ImmutableSet
							.builder();
					final Iterator<?> iterator = selection.iterator();
					while (iterator.hasNext()) {
						@SuppressWarnings("unchecked")
						final Map.Entry<Integer, Map<Witness, Set<Token>>> entry = (Map.Entry<Integer, Map<Witness, Set<Token>>>) iterator
						.next();
						builder.addAll(entry.getValue().get(witness));
					}
					final ImmutableSet<Token> tokens = builder.build();
					WitnessTransfer.getInstance().setTokens(tokens);
					event.data = tokens;
					// System.err.println(tokens);
				}

				@Override
				public void dragStart(final DragSourceEvent event) {
					System.err.println("Drag started"); //$NON-NLS-1$
					super.dragStart(event);
					dragSetData(event);
				}

			});

			if (collationSetEditor != null
					&& collationSetEditor.getCollationResults() != null)
				getViewer().setInput(collationSetEditor
						.getCollationResults());
		}
		control = container;
	}

	@Override
	public Control getControl() {
		return control;
	}

	@Override
	public void init(final IPageSite pageSite) {
		super.init(pageSite);
		createActions();
		initializeToolBar();
		initializeMenu();
	}

	@Override
	public void setFocus() {
		// Set the focus
	}

	/**
	 * Create the actions.
	 */
	private void createActions() {
		// Create the actions
	}

	/**
	 * Initialize the toolbar.
	 */
	private void initializeToolBar() {
		final IToolBarManager toolbarManager = getSite().getActionBars()
				.getToolBarManager();
	}

	/**
	 * Initialize the menu.
	 */
	private void initializeMenu() {
		final IMenuManager menuManager = getSite().getActionBars().getMenuManager();
	}

	@Override
	protected GridTableViewer getViewer() {
		return gridTableViewer;
	}


	protected void setViewer(final GridTableViewer gridTableViewer) {
		this.gridTableViewer = gridTableViewer;
	}

}

final class CollationResultsContentProvider implements
IStructuredContentProvider, ICollationResultsListener {
	private GridTableViewer viewer;
	private VariantGraph variantGraph;
	private RowSortedTable<Integer, Witness, Set<Token>> alignmentTable;
	private Set<VariantGraphTransposition> transpositions;
	private final Map<GridViewerColumn, Witness> columns = Maps
			.newLinkedHashMap();

	public CollationResultsContentProvider() {
		// TODO Auto-generated constructor stub
	}

	@Override
	public void dispose() {
	}

	@Override
	public void inputChanged(final Viewer viewer, final Object oldInput,
			final Object newInput) {
		this.viewer = (GridTableViewer) viewer;

		if (oldInput instanceof CollationResults)
			((CollationResults) oldInput).removeListener(this);

		if (newInput != null && newInput instanceof CollationResults) {

			final Iterator<GridViewerColumn> iterator = columns.keySet()
					.iterator();
			while (iterator.hasNext()) {
				final GridViewerColumn col = iterator.next();
				col.getColumn().dispose();
				iterator.remove();
			}

			final CollationResults collationResults = (CollationResults) newInput;
			variantGraph = collationResults.getVariantGraph();
			final Transaction transaction = variantGraph.getDatabase()
					.beginTx();
			try {
				alignmentTable = variantGraph.toTable();
				transpositions = variantGraph.transpositions();
				transaction.success();
			} finally {
				transaction.finish();
			}
			final int witnessCount = alignmentTable.columnKeySet().size();
			final int columnWidth = Math.max(50, viewer.getControl()
					.getBounds().width / witnessCount);
			for (final Witness witness : alignmentTable.columnKeySet()) {
				final GridViewerColumn viewerColumn = new GridViewerColumn(
						this.viewer, SWT.NONE);
				columns.put(viewerColumn, witness);
				viewerColumn
				.setLabelProvider(new DelegatingStyledCellLabelProvider(
						new WitnessLabelProvider(collationResults,
								witness)));
				viewerColumn.getColumn().setText(witness.getSigil());
				viewerColumn.getColumn().setWidth(columnWidth);
			}
			collationResults.addListener(this);
		}
	}

	@Override
	public Object[] getElements(final Object inputElement) {
		if (inputElement instanceof CollationResults
				&& ((CollationResults) inputElement).getVariantGraph() == variantGraph)
			return alignmentTable.rowMap().entrySet().toArray();
		return new Object[0];
	}

	public Witness getWitness(final GridColumn focusColumn) {
		for (final GridViewerColumn col : columns.keySet())
			if (col.getColumn().equals(focusColumn))
				return columns.get(col);
		return null;
	}

	@Override
	public void focusChanged(final CollationResults results) {
		viewer.update(getElements(results), null);
	}
}
